#include "xbee_conf.h"
#include "xbeemicrolib.h"
#include "xbee_uart.h"

#if 1

#ifdef _DEBUG
#include <stdio.h>
#include <windows.h>
#endif

#ifndef NULL
#define NULL (void*)0
#endif


#define XBEE_TX_BUFFER_EMPTY 0
#define XBEE_TX_BUFFER_WAIT_AT_REPLY 1
#define XBEE_TX_BUFFER_WAIT_TRANSMIT_REPLY 2


#define XBEE_AT_OK                0x00
#define XBEE_AT_ERROR             0x01
#define XBEE_AT_INVALID_COMMAND   0x02
#define XBEE_AT_INVALID_PARAMETER 0x03

#define XBEE_MODEM_STAUS_HARDWARE_RESET        0
#define XBEE_MODEM_STAUS_WATCHDOG_RESET        1
#define XBEE_MODEM_STAUS_JOIN_NETWORK          2
#define XBEE_MODEM_STAUS_DISASSOCIATED         3
#define XBEE_MODEM_STAUS_COORDINATOR_STARTED   4

/**
 * Packet Instructions
 */
#define XBEE_INST_AT_COMMAND              0x08
#define XBEE_INST_AT_COMMAND_BUFFER_PARAM 0x09
#define XBEE_INST_ZIGBEE_TRANSMIT_REQUEST 0x10
#define XBEE_INST_EXPLICIT_ZIGBEE_COMMAND 0x11
#define XBEE_INST_REMOTE_COMMAND_REQUEST  0x17
#define XBEE_INST_CREATE_SOURCE_ROUTE     0x21
#define XBEE_INST_AT_COMMAND_RESPONCE     0x88
#define XBEE_INST_MODEM_STATUS            0x8A
#define XBEE_INST_TRANSMIT_STATUS         0x8B
#define XBEE_INST_RECEIVE_PACKET          0x90
#define XBEE_INST_EXPLICIT_RX_INDICATOR   0x91
#define XBEE_INST_ZIGBEE_IO_DATA_SAMPLE   0x92
#define XBEE_INST_XBEE_SENSOR_READ        0x94
#define XBEE_INST_NODE_IDENTIFICATION     0x95
#define XBEE_INST_REMOTE_COMMAND_RESPONCE 0x97
#define XBEE_INST_OVER_THE_AIR_UPDATE     0xA0
#define XBEE_INST_ROUTE_RECORD_INDICATOR  0xA1




/********************/
/* Internal values  */
/********************/
static int mode = XBEE_MODE_STARTUP;
static volatile int softwareReset = 0;
static unsigned char at_command_responce = 0;
static unsigned char tx_command_responce = 0;
static unsigned char packet_id = 1;

const unsigned short wait_count = 2000;


/*************/
/* Tx Buffer */
/*************/
unsigned char tx_buffer_status = XBEE_TX_BUFFER_EMPTY;
unsigned char tx_packet_buffer[128];
unsigned short tx_packet_size;
unsigned short tx_try_count = -1;
unsigned short tx_try_count_buffer;
unsigned short tx_try_packet_id;

unsigned char XBee_getTxBufferStatus() {return tx_buffer_status;}


/**************/
/* Rx Buffer  */
/**************/
unsigned char  rx_packet_buffer[128];
unsigned short rx_packet_size;



int XBee_getMode() {return mode;}

static int XBee_sendATCommand(const char* command, const char* arg, const unsigned short size);
static unsigned char XBee_waitATCommandDone(void);
static void XBee_waitSoftReset(void);

static void uart_putc_escaped(unsigned char c);
static int uart_getc_escaped(unsigned char* c);
static void XBee_sendPacket();
static int XBee_receivePacket();

static void onReplyATDN();




static int on_rx_initialized = 0;

void(*XBee_onRxPacket)(const char* addr64, const char* addr16, const char* dat, const unsigned short size);

void XBee_registerOnRxPacket(void(*onRxPacket)(const char* addr64, const char* addr16, const char* dat, const unsigned short size)) {
	XBee_onRxPacket = onRxPacket;
	on_rx_initialized = 1;
}



/**
 * @brief Initialization Function
 *  Here, settings below are configured automatically.
 *  1. AP=2, 
 *  2. NI (Network Identifier String)
 *  3. SC (Scan PAN Mask)
 *  4. PANID (64bit PAN ID)
 */
void XBee_setup(void) {
	
	mode = XBEE_MODE_STARTUP;
	uart_setup();
	
	XBee_sendATCommand("FR", NULL, 0);
	XBee_waitSoftReset();

	do {
		XBee_sendATCommand("AP", &AP, 1);
	} while(XBee_waitATCommandDone() != XBEE_AT_OK);

	XBee_sendATCommand("SC", scan_channel_mask, 2);
	XBee_waitATCommandDone();

	XBee_sendATCommand("ID", pan64bitID, 8);
	XBee_waitATCommandDone();

	XBee_sendATCommand("NI", networkIdentifierString, NI_STRING_LENGTH);
	XBee_waitATCommandDone();
	
}


void XBee_onRxModemStatus(void) {
#ifdef _DEBUG
	int i;
	printf("Modem:");
	for(i = 0;i < rx_packet_size;i++) {
		printf("%x,", rx_packet_buffer[i]);
	}
	putchar('\n');
#endif
	switch(rx_packet_buffer[1]) {
		case XBEE_MODEM_STAUS_HARDWARE_RESET:
			break;
		case XBEE_MODEM_STAUS_WATCHDOG_RESET:
			softwareReset = 1;
			mode = XBEE_MODE_STARTUP;
			break;
		case XBEE_MODEM_STAUS_JOIN_NETWORK:
			mode = XBEE_MODE_INITIALIZED;
			break;
		case XBEE_MODEM_STAUS_DISASSOCIATED:
			break;
		case XBEE_MODEM_STAUS_COORDINATOR_STARTED:
			mode = XBEE_MODE_INITIALIZED;
			break;
	}
}



void XBee_onRxTransmitStatus(void) {
	tx_command_responce = rx_packet_buffer[5];
	tx_buffer_status = XBEE_TX_BUFFER_EMPTY;
#ifdef _DEBUG
	if(rx_packet_buffer[5] == 0x00) {
		int i;
		printf("Transmit Command (%d) Succeeded.\n", rx_packet_buffer[1]);
		for(i = 0;i < rx_packet_size;i++) {
			printf("%x,", rx_packet_buffer[i]);
		}
		putchar('\n');
	} else {
		printf("Transmit Command (%d) Failed.\n", rx_packet_buffer[1]);
	}
#endif
}


int XBee_polling(void) {
	int ret, i;
	int timeout = 1;

	// đM
	if(tx_try_count_buffer > 0 && tx_buffer_status != XBEE_TX_BUFFER_EMPTY) {
		tx_try_count_buffer--;
		if(tx_try_count_buffer == 0) {
			XBee_sendPacket();
		}
	}

	ret = XBee_receivePacket();
	if(ret == ERR_RX_EMPTY) {
		return ERR_TIMEOUT;
	}
	switch(ret) {
		case XBEE_INST_MODEM_STATUS:
			XBee_onRxModemStatus();
			break;
		case XBEE_INST_TRANSMIT_STATUS:
			XBee_onRxTransmitStatus();
			break;	
		case XBEE_INST_RECEIVE_PACKET:
			if(on_rx_initialized) {
				XBee_onRxPacket(rx_packet_buffer+1, rx_packet_buffer+9, rx_packet_buffer+12, rx_packet_size-12);
			}
			break;
		case XBEE_INST_AT_COMMAND_RESPONCE:
			at_command_responce = rx_packet_buffer[4];
			tx_buffer_status = XBEE_TX_BUFFER_EMPTY;
#ifdef _DEBUG
			printf("AT Command (%c%c) =%d.\n", rx_packet_buffer[2], rx_packet_buffer[3], rx_packet_buffer[4]);
#endif
			if(rx_packet_buffer[2] == 'D' && rx_packet_buffer[3] == 'N') onReplyATDN();
			break;
		default:
			return 0;
			break;
	}
	return ret;
}


static void XBee_waitSoftReset(void) {
	while(!softwareReset) {
		XBee_polling();
	}
	softwareReset = 0;
}


#define XBEE_BLOADCAST_RADIUS_MAX 0x00
#define XBEE_DEFAULT_TXOPTION     0x00
/**
 * @brief packet transmit
 *
 * @param pDst destination host record (this should be ready for transmit. see XBee_resolve)
 * @param dat  data buffer
 * @param size size of data
 * @return zero if success. ERR_UART_BUSY (Tx Buffer is busy). ERR_INVALID_HOST (host record is not ready for use)
 */
int XBee_transmit(const XBeeHostRecord* pDst, const unsigned char* dat, const unsigned char size) {
	unsigned char i;
	unsigned short cnt = 0;
	if(tx_buffer_status != XBEE_TX_BUFFER_EMPTY) {
		return ERR_UART_BUSY;
	}
	if(pDst->status != XBEE_HOST_KNOWN) {
		return ERR_INVALID_HOST;
	}

	tx_packet_buffer[0] = XBEE_INST_ZIGBEE_TRANSMIT_REQUEST; // Transmit Command
	tx_packet_buffer[1] = packet_id;
	tx_packet_buffer[2] = pDst->addr64[0];
	tx_packet_buffer[3] = pDst->addr64[1];
	tx_packet_buffer[4] = pDst->addr64[2];
	tx_packet_buffer[5] = pDst->addr64[3];
	tx_packet_buffer[6] = pDst->addr64[4];
	tx_packet_buffer[7] = pDst->addr64[5];
	tx_packet_buffer[8] = pDst->addr64[6];
	tx_packet_buffer[9] = pDst->addr64[7];
	tx_packet_buffer[10] = pDst->addr16[0];
	tx_packet_buffer[11] = pDst->addr16[1];
	tx_packet_buffer[12] = XBEE_BLOADCAST_RADIUS_MAX;
	tx_packet_buffer[13] = XBEE_DEFAULT_TXOPTION;
	for(i = 0;i < size;i++) {
		tx_packet_buffer[14+i] = dat[i];
	}
	tx_packet_size =  size+14;
	tx_try_count = 10000;
	tx_buffer_status = XBEE_TX_BUFFER_WAIT_TRANSMIT_REPLY;

	XBee_sendPacket();
	return 0;
}



/**
 * @brief wait transmit is done
 *
 * @return success:XBEE_DELIVERY_STATUS_SUCCESS, error see XBEE_DELIVERY_STATUS_
 */
unsigned char XBee_waitTransmitCommandDone(void) {
	while(XBee_getTxBufferStatus() == XBEE_TX_BUFFER_WAIT_TRANSMIT_REPLY) {
		XBee_polling();
	}
	return tx_command_responce;
}


/***************************************/
/* Send AT Command                     */
/* Command is immediately applied.     */
/***************************************/
int XBee_sendATCommand(const char* command, const char* arg, const unsigned short size) {
	unsigned short i;
	if(tx_buffer_status != XBEE_TX_BUFFER_EMPTY) {
		return ERR_TRANSMIT_BUSY;
	}

	tx_packet_buffer[0] = XBEE_INST_AT_COMMAND; // AT command 
	tx_packet_buffer[1] = packet_id; // Frame ID
	tx_packet_buffer[2] = command[0];
	tx_packet_buffer[3] = command[1];
	for(i = 0;i < size;i++) {
		tx_packet_buffer[4 + i] = arg[i];
	}
	tx_packet_size = 4 + size;
	tx_try_count = 100;
	tx_buffer_status = XBEE_TX_BUFFER_WAIT_AT_REPLY;
	XBee_sendPacket(tx_packet_buffer, size+4);
	return 0;
}

unsigned char XBee_waitATCommandDone(void) {
	while(XBee_getTxBufferStatus() == XBEE_TX_BUFFER_WAIT_AT_REPLY) {
		XBee_polling();
	}
	return at_command_responce;
}


/**************************/
/* Host Address Resolving */
/**************************/

/***************/
/* Host Buffer */
/***************/
static XBeeHostRecord *pNowResolving = NULL;

/**
 * @brief resolve network address from node identifier string
 *
 * @param pHost host record.
 * @return zero if success.
 */
int XBee_resolve(XBeeHostRecord* pHost) {
	int i = 0;
	pNowResolving = pHost;
	while(1) {
		if(pHost->networkIdentifierString[i] == '\0') {
			break;
		}
		i++;
	}
	return XBee_sendATCommand("DN", pHost->networkIdentifierString, i);
}


static void onReplyATDN(void) {
	if(pNowResolving == NULL) return;

	if(at_command_responce == 0x00) {
		pNowResolving->addr16[0] = rx_packet_buffer[5];
		pNowResolving->addr16[1] = rx_packet_buffer[6];
		pNowResolving->addr64[0] = rx_packet_buffer[7];
		pNowResolving->addr64[1] = rx_packet_buffer[8];
		pNowResolving->addr64[2] = rx_packet_buffer[9];
		pNowResolving->addr64[3] = rx_packet_buffer[10];
		pNowResolving->addr64[4] = rx_packet_buffer[11];
		pNowResolving->addr64[5] = rx_packet_buffer[12];
		pNowResolving->addr64[6] = rx_packet_buffer[13];
		pNowResolving->addr64[7] = rx_packet_buffer[14];
		pNowResolving->status    = XBEE_HOST_KNOWN;
#ifdef _DEBUG
		printf("Resolve (%s)\n", pNowResolving->networkIdentifierString);
		printf(" <%x><%x> \n", pNowResolving->addr16[0],pNowResolving->addr16[1]);
		printf(" <%x><%x><%x><%x><%x><%x><%x><%x> \n", 
			pNowResolving->addr64[0],pNowResolving->addr64[1],pNowResolving->addr64[2],pNowResolving->addr64[3],
			pNowResolving->addr64[4],pNowResolving->addr64[5],pNowResolving->addr64[6],pNowResolving->addr64[7]);
#endif
	}
	pNowResolving = NULL;
}

/**
 * @brief wait until resolve is done
 *
 * @return zero if success.
 */
unsigned char XBee_waitResolveDone(void) {
	while(XBee_getTxBufferStatus() == XBEE_TX_BUFFER_WAIT_AT_REPLY) {
		XBee_polling();
	}
	return at_command_responce;
}

/************************************/
/* Lowest Level Accessing functions */
/************************************/


#define XBEE_API_FRAME_DELIMITER         0x7E
#define XBEE_API_FRAME_XON               0x11
#define XBEE_API_FRAME_XOFF              0x13
#define XBEE_API_FRAME_ESCAPE_CHARACTER  0x7D
#define XBEE_API_FRAME_ESCAPE_FLAG       0x20

#define XBEE_MAXIMUM_PAYLOADSIZE 0x08
#define XBEE_TRANSMITPACKET_HEADER_LENGTH (16)





/**
 * @brief send packet in tx buffer
 */
static void XBee_sendPacket() {
	unsigned short i;
	unsigned char sum = 0;
	unsigned char c;
	uart_putc(XBEE_API_FRAME_DELIMITER); // Do not escape here.
	c = tx_packet_size >> 8;
	uart_putc_escaped(c);
	c = tx_packet_size & 0xff;
	uart_putc_escaped(c);
	for(i = 0;i < tx_packet_size;i++) {
		uart_putc_escaped(tx_packet_buffer[i]);
		sum += tx_packet_buffer[i];
	}
	uart_putc_escaped(0xff-sum);

	tx_try_count_buffer = tx_try_count;
	tx_try_packet_id = packet_id;

	packet_id++;
	if(packet_id == 0) packet_id = 1;
}

/**
 * @brief receive packet to rx buffer.
 *
 */
static int XBee_receivePacket() {
	unsigned short i = 0;
	unsigned char c;
	unsigned char sum = 0;
	if(!uart_isReceived()) {
		return ERR_RX_EMPTY;
	}

	uart_setBlockingPolicy(1);
	uart_getc_escaped(&c); // DELIMITER
	if(c != XBEE_API_FRAME_DELIMITER) return ERR_INVALID;
	uart_getc_escaped(&c); // SIZE (MSB)
	rx_packet_size = (unsigned short)c << 8;
	uart_getc_escaped(&c);
	rx_packet_size |= c; // SIZE (LSB)
	for(i = 0;i < rx_packet_size;i++) {
		uart_getc_escaped(rx_packet_buffer+i);
		sum += rx_packet_buffer[i];
	}
	uart_getc_escaped(&c);
	uart_setBlockingPolicy(0);
	if(c != (0xff-sum)) {return ERR_CHECKSUM;}
	return rx_packet_buffer[0];
}




/**************************************/
/* UART (escape character considered) */
/**************************************/

/**
 * @brief escaped uart put character
 *
 * @param c character output
 */
static void uart_putc_escaped(unsigned char c) {
	if(c == XBEE_API_FRAME_DELIMITER ||
		c == XBEE_API_FRAME_ESCAPE_CHARACTER ||
		c == XBEE_API_FRAME_XON ||
		c == XBEE_API_FRAME_XOFF) {
		uart_putc(XBEE_API_FRAME_ESCAPE_CHARACTER);
		uart_putc(XBEE_API_FRAME_ESCAPE_FLAG ^ c);
	} else {
		uart_putc(c);
	}
}

/**
 * @brief escaped uart get character
 *
 * @param c character input
 * @return zero if success
 */
static int uart_getc_escaped(unsigned char* c) {
	int ret = uart_getc(c);
	if(ret != 0) return ret;
	if(*c == XBEE_API_FRAME_ESCAPE_CHARACTER) {
		ret = uart_getc(c);
		if(ret != 0) return ret;
		*c = *c ^ XBEE_API_FRAME_ESCAPE_FLAG;
	}
	return ret;
}
#endif